"""
    Flowblade Movie Editor is a nonlinear video editor.
    Copyright 2012 Janne Liljeblad.

    This file is part of Flowblade Movie Editor <https://github.com/jliljebl/flowblade/>.

    Flowblade Movie Editor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Flowblade Movie Editor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Flowblade Movie Editor.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
Module handles keyevents.
"""

from gi.repository import Gdk, Gtk

import re

import appconsts
import clipeffectseditor
import compositeeditor
import compositormodes
import glassbuttons
import gui
import editorpersistance
import editorstate
from editorstate import current_sequence
from editorstate import PLAYER
from editorstate import timeline_visible
import keyframeeditor
import kftoolmode
import medialog
import menuactions
import modesetting
import monitorevent
import movemodes
import multitrimmode
import shortcuts
import shortcutsquickeffects
import syncsplitevent
import render
import targetactions
import tlineaction
import tlinewidgets
import tlineypage
import trackaction
import trimmodes
import updater
import projectaction
import workflow

# TODO: We should consider integrating some parts of this with targetactions.py
# TODO:
# TODO: As of this writing, targetactions.py has a superset of targetable
# TODO: actions, as compared to keyevents.py, totally separate from any keyboard
# TODO: event handling. There are a few new named target actions in there that
# TODO: aren't available in here. There is also currently a lot code duplication
# TODO: between the two modules. See targetactions.py for more details.
# TODO:
# TODO: At a minimum, if you add or modify any of the key actions in here,
# TODO: please consider updating targetactions.py as well. Right now there
# TODO: is a lot of duplication between these modules, and often a change
# TODO: in one would warrant a change in the other.
# TODO:
# TODO: keyevents.py is all about handling key presses from the keyboard, and
# TODO: routing those events to trigger actions in various parts of the program.
# TODO:
# TODO: targetactions.py is basically a bunch of zero-argument functions with
# TODO: names based on the shortcut key names found here. It was created as part
# TODO: of the USB HID work, so that USB jog/shuttle devices could have their
# TODO: buttons target various actions within the program, without requiring
# TODO: each USB driver to directly make connections to a dozen different parts
# TODO: of the program to control it.
# TODO:
# TODO: So now we have two collections of shortcut key names which map to
# TODO: basically the same actions, but in a different way. I originally wanted
# TODO: to just use keyevents.py as the target for the USB driver actions, but
# TODO: couldn't use it directly since this module is intertwined with the
# TODO: main computer keyboard and its events.
# TODO:
# TODO: For now, I have integrated the new command targets from
# TODO: targetactions.py into keyevents.py, both for completeness, and also as
# TODO: a proof of concept as to how we might migrate some of the other code
# TODO: in here over to call targetactions.py
# TODO:
# TODO:   -- Nathan Rosenquist (@ratherlargerobot)
# TODO:      Feb 2022

# ------------------------------------- keyboard events
def key_down(widget, event):
    
    #print (_get_shortcut_action(event))
    
    """
    Global key press listener.
    """
    # Handle ESCAPE.
    if event.keyval == Gdk.KEY_Escape:
        if editorstate.current_is_move_mode() == False:
            modesetting.set_default_edit_mode()
            return True
        elif gui.big_tc.get_visible_child_name() == "BigTCEntry":
            gui.big_tc.set_visible_child_name("BigTCDisplay")
            return True
    
    # Make Home and End work on name entry widget.
    # TODO: See which other components could benefit from this check.
    if render.widgets.file_panel.movie_name.has_focus():
        return False

    # If timeline widgets are in focus timeline keyevents are available.
    if _timeline_has_focus():
        was_handled = _handle_tline_key_event(event)
        if was_handled:
            # Stop widget focus from travelling if arrow key pressed for next frame
            # by stopping signal.
            gui.editor_window.window.emit_stop_by_name("key_press_event")
        return was_handled
    
    # Insert shortcut keys need more focus then timeline shortcuts.
    # these may already have been handled in timeline focus events.
    was_handled = _handle_extended_monitor_focus_events(event)
    if was_handled:
        # Stop event handling here
        return True

    was_handled = _handle_configurable_global_events(event)
    if was_handled:
        return True

    # Pressing timeline button obviously leaves user expecting
    # to have focus in timeline.
    if gui.monitor_switch.widget.has_focus() and timeline_visible():
        was_handled = _handle_tline_key_event(event)
        return was_handled

    #  Handle non-timeline delete.
    if event.keyval == Gdk.KEY_Delete:
        return _handle_delete()

    # Select all with CTRL + A in media panel.
    if event.keyval == Gdk.KEY_a:
        if (event.get_state() & Gdk.ModifierType.CONTROL_MASK):
            if gui.media_list_view.widget.has_focus() or gui.media_list_view.widget.get_focus_child() != None: 
                gui.media_list_view.select_all()
                return True

    if event.keyval == Gdk.KEY_F11:
        menuactions.toggle_fullscreen()
        return True

    # Key event was not handled here.
    #print(_get_shortcut_action(event), " not handled")
    return False
    
def _timeline_has_focus():
    if gui.editor_window.tool_selector != None and gui.editor_window.tool_selector.widget.has_focus():
        return True
    
    if(gui.tline_canvas.widget.has_focus()
       or gui.tline_column.widget.has_focus()
       or (gui.pos_bar.widget.has_focus() and timeline_visible())
       or gui.tline_scale.widget.has_focus()
       or glassbuttons.focus_group_has_focus(glassbuttons.DEFAULT_FOCUS_GROUP)):
        return True

    return False
    
def _handle_tline_key_event(event):
    """
    This is called when timeline widgets have focus and key is pressed.
    Returns True for handled key presses to stop those
    keyevents from going forward.
    """
    tool_was_selected = workflow.tline_tool_keyboard_selected(event)
    if tool_was_selected == True:
        return True

    was_handled = shortcutsquickeffects.maybe_do_quick_shortcut_filter_add(event)
    if was_handled == True:
        return True

    return False

def _handle_extended_monitor_focus_events(event):
    tool_was_selected = workflow.tline_tool_keyboard_selected(event)
    if tool_was_selected == True:
        return True

    return False
        
# Apr-2017 - SvdB
def _get_shortcut_action(event):
    return shortcuts.get_shortcut_action(event)

def _handle_configurable_global_events(event):
    action = _get_shortcut_action(event)
    if action == 'tline_page_up':
        tlineypage.page_up_key()
        return True
    if action == 'tline_page_down':
        tlineypage.page_down_key()
        return True
    if action == 'open_next':
        projectaction.open_next_media_item_in_monitor()
        return True
    if action == 'open_prev':
        projectaction.open_prev_media_item_in_monitor()
        return True
    if action == "append_from_bin":
        if gui.media_list_view.widget.has_focus() or gui.media_list_view.widget.get_focus_child() != None: 
            projectaction.append_selected_media_clips_into_timeline()
            return True
    if action == "move_media":
        gui.media_list_view.init_move()
    if action == 'monitor_show_video':
        tlineaction.set_monitor_display_mode(appconsts.PROGRAM_OUT_MODE)
        return True
    if action == 'monitor_show_scope':
        tlineaction.set_monitor_display_mode(appconsts.VECTORSCOPE_MODE)
        return True
    if action == 'monitor_show_rgb':
        tlineaction.set_monitor_display_mode(appconsts.RGB_PARADE_MODE)
        return True

    return False

def _handle_delete():
    # Delete media file
    if gui.media_list_view.widget.get_focus_child() != None:
        projectaction.delete_media_files()
        return True

    return False



"""
    Flowblade Movie Editor is a nonlinear video editor.
    Copyright 2025 Janne Liljeblad.

    This file is part of Flowblade Movie Editor <https://github.com/jliljebl/flowblade/>.

    Flowblade Movie Editor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Flowblade Movie Editor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Flowblade Movie Editor.  If not, see <http://www.gnu.org/licenses/>.
"""
from gi.repository import Gdk

import appconsts
import clipeffectseditor
import compositeeditor
import compositormodes
import editorpersistance
import editorstate
from editorstate import PLAYER
from editorstate import current_sequence
import gui
import keyframeeditor
import kftoolmode
import medialog
import monitorevent
import movemodes
import multitrimmode
import projectaction
import shortcuts
import syncsplitevent
import tlineaction
import tlinewidgets
import trackaction
import trimmodes
import updater

# Widget names for action target setting and look up.
SEQUENCE_LIST_VIEW = "sequencelistview"
BIN_LIST_VIEW = "binlistview"
LOG_LIST_VIEW = "loglistview"
TLINE_CANVAS = "tlinecanvas"
TLINE_TRACK_COLUMN = "tlinetrackcolumn"
TLINE_SCALE = "tlinescale"
POS_BAR = "posbar"
TLINE_LEFT_CORNER = "tlineleftcorner"
TLINE_SCROLL = "tlinescroll"
TLINE_MONITOR_DISPLAY = "tlinemonitordisplay"
TLINE_ALL = [TLINE_CANVAS, TLINE_TRACK_COLUMN, TLINE_SCALE, TLINE_LEFT_CORNER, TLINE_SCROLL, TLINE_MONITOR_DISPLAY]
MONITOR_SWITCH = "monitorswitch"
#MONITOR_WAVEFORM_DISPLAY = "monitorwaveformdisplay"
MONITOR_ALL = [MONITOR_SWITCH, POS_BAR]
MIDDLE_BUTTONS_1 = "midbar_b_1"
MIDDLE_BUTTONS_2 = "midbar_b_2"
MIDDLE_BUTTONS_3 = "midbar_b_3"
MIDDLE_BUTTONS_4 = "midbar_b_4"
MIDDLE_BUTTONS_5 = "midbar_b_5"
MIDDLE_BUTTONS_6 = "midbar_b_6"
MIDDLEBAR_BUTTONS = [MIDDLE_BUTTONS_1, MIDDLE_BUTTONS_2, MIDDLE_BUTTONS_3, MIDDLE_BUTTONS_4, MIDDLE_BUTTONS_5, MIDDLE_BUTTONS_6]
TLINE_MONITOR_ALL = TLINE_ALL + MONITOR_ALL + MIDDLEBAR_BUTTONS

# widget name -> widget object
_widgets = {}
# widget -> ShortCutController object
_controllers = {}


def init():
    # Build _widgets data struct with live objects.
    global _widgets
    _widgets = {TLINE_CANVAS: gui.tline_canvas.widget,
                TLINE_TRACK_COLUMN: gui.tline_column.widget,
                TLINE_SCALE: gui.tline_scale.widget,
                POS_BAR: gui.pos_bar.widget,
                TLINE_LEFT_CORNER: gui.tline_left_corner.widget,
                TLINE_SCROLL: gui.tline_scroll,
                TLINE_MONITOR_DISPLAY: gui.tline_display,
                MONITOR_SWITCH: gui.monitor_switch.widget,
                SEQUENCE_LIST_VIEW: gui.sequence_list_view,
                BIN_LIST_VIEW: gui.bin_list_view, 
                LOG_LIST_VIEW: gui.editor_window.media_log_events_list_view,
                MIDDLE_BUTTONS_1: gui.editor_window.zoom_buttons.widget,
                MIDDLE_BUTTONS_2: gui.editor_window.edit_buttons.widget,
                MIDDLE_BUTTONS_3: gui.editor_window.edit_buttons_3.widget, 
                MIDDLE_BUTTONS_4: gui.editor_window.edit_buttons_2.widget, 
                MIDDLE_BUTTONS_5: gui.editor_window.monitor_insert_buttons.widget, 
                MIDDLE_BUTTONS_6: gui.editor_window.undo_redo.widget}

    # Create actions
    _create_action("mark_in", monitorevent.mark_in_pressed, TLINE_MONITOR_ALL)
    _create_action("to_mark_in", monitorevent.to_mark_in_pressed, TLINE_MONITOR_ALL)
    _create_action("mark_out", monitorevent.mark_out_pressed, TLINE_MONITOR_ALL)
    _create_action("to_mark_out", monitorevent.to_mark_out_pressed, TLINE_MONITOR_ALL)
    _create_action("clear_io_marks", monitorevent.marks_clear_pressed, TLINE_MONITOR_ALL)
    _create_action("zoom_out", updater.zoom_out, TLINE_MONITOR_ALL)
    _create_action("zoom_in", updater.zoom_in, TLINE_MONITOR_ALL)
    _create_action("switch_monitor", updater.switch_monitor_display, TLINE_MONITOR_ALL)
    _create_action("add_marker", tlineaction.add_marker, TLINE_MONITOR_ALL)
    _create_action("cut", tlineaction.cut_pressed, TLINE_ALL)
    _create_action("cut_all", tlineaction.cut_all_pressed, TLINE_ALL)
    _create_action("log_range", medialog.log_range_clicked, TLINE_MONITOR_ALL)
    _create_action("set_length", tlineaction.set_length_from_keyevent, TLINE_ALL)
    _create_action("insert", tlineaction.insert_button_pressed, TLINE_MONITOR_ALL)
    _create_action("append", tlineaction.append_button_pressed, TLINE_MONITOR_ALL)
    _create_action("3_point_overwrite", tlineaction.three_point_overwrite_pressed, TLINE_MONITOR_ALL)
    _create_action("overwrite_range", tlineaction.range_overwrite_pressed, TLINE_MONITOR_ALL)
    _create_action("trim_start", tlineaction.trim_start_pressed, TLINE_ALL)
    _create_action("trim_end", tlineaction.trim_end_pressed, TLINE_ALL)
    _create_action("toggle_audio_mute", tlineaction.mute_clip_from_keyevent, TLINE_ALL)
    _create_action("nudge_back", lambda: movemodes.nudge_selection(-1), TLINE_ALL)
    _create_action("nudge_forward", lambda: movemodes.nudge_selection(1), TLINE_ALL)
    _create_action("nudge_back_10", lambda: movemodes.nudge_selection(-10), TLINE_ALL)
    _create_action("nudge_forward_10", lambda: movemodes.nudge_selection(10), TLINE_ALL)
    _create_action("select_next", monitorevent.select_next_clip_for_filter_edit, TLINE_MONITOR_ALL)
    _create_action("select_prev", monitorevent.select_prev_clip_for_filter_edit, TLINE_MONITOR_ALL)
    _create_action("toggle_track_output", trackaction.toggle_track_output, TLINE_ALL)
    _create_action("split_selected", tlineaction.split_audio_synched_button_pressed, TLINE_ALL)
    _create_action("set_sync_relation", syncsplitevent.init_select_master_clip_from_keyevent, TLINE_ALL)
    _create_action("clear_sync_relation", syncsplitevent.clear_sync_relation_from_keyevent, TLINE_ALL)
    _create_action("slower", monitorevent.j_pressed, TLINE_MONITOR_ALL)
    _create_action("stop", monitorevent.k_pressed, TLINE_MONITOR_ALL)
    _create_action("faster", monitorevent.l_pressed, TLINE_MONITOR_ALL)
    _create_action("play_pause", _play_pause_action, TLINE_MONITOR_ALL)
    _create_action("prev_frame", _prev_frame_action, TLINE_MONITOR_ALL, True)
    _create_action("next_frame", _next_frame_action, TLINE_MONITOR_ALL, True)
    _create_action("next_cut", _next_cut_action, TLINE_MONITOR_ALL)
    _create_action("prev_cut", _prev_cut_action, TLINE_MONITOR_ALL)
    _create_action("enter_edit", _enter_edit_action, TLINE_ALL)
    _create_action("delete", _tline_delete_action, TLINE_ALL)
    _create_action("lift", tlineaction.lift_button_pressed, TLINE_ALL) 
    _create_action("play_pause_loop_marks", _play_pause_loop_marks_action, TLINE_MONITOR_ALL)
    _create_action("to_start", _to_start_action, TLINE_MONITOR_ALL)
    _create_action("to_end", _to_end_action, TLINE_MONITOR_ALL)
    _create_action("delete", _delete_sequence_action, [SEQUENCE_LIST_VIEW])
    _create_action("delete", _delete_bin_action, [BIN_LIST_VIEW])
    _create_action("delete", _delete_log_action, [LOG_LIST_VIEW])
    _create_action("append_from_bin", projectaction.append_selected_media_clips_into_timeline, TLINE_MONITOR_ALL)
    _create_action("resync", tlineaction.resync_track_button_pressed, TLINE_MONITOR_ALL) # Currently menu global gets to this firsts.
    _create_action("monitor_show_video", lambda: tlineaction.set_monitor_display_mode(appconsts.PROGRAM_OUT_MODE), TLINE_MONITOR_ALL)
    _create_action("monitor_show_scope", lambda: tlineaction.set_monitor_display_mode(appconsts.VECTORSCOPE_MODE), TLINE_MONITOR_ALL)
    _create_action("monitor_show_rgb", lambda: tlineaction.set_monitor_display_mode(appconsts.RGB_PARADE_MODE), TLINE_MONITOR_ALL)
    _create_action("mark_selection_range", monitorevent.mark_selection_range_pressed, TLINE_MONITOR_ALL)

def _create_action(action, press_func, widget_list, pass_event=False):
    for widget_id in widget_list:
        widget = _widgets[widget_id]
        #print(widget_id)
        if widget in _controllers:
            _controllers[widget].add_shortcut(action, press_func, pass_event)
        else:
            #print("ShortCutController")
            _controllers[widget] = ShortCutController(widget)
            _controllers[widget].add_shortcut(action, press_func, pass_event)


class ShortCutController:
    def __init__(self, widget):
        self.shortcuts = {}
        self.widget = widget
        self.widget.connect("key-press-event", lambda w, e: self._short_cut_handler(e))
    
    def add_shortcut(self, action, press_func, pass_event):
        self.shortcuts[action] = (press_func, pass_event)
    
    def _short_cut_handler(self, event):
        action = shortcuts.get_shortcut_action(event)
        #print("ShortCutController: ", action)
        try:
            press_func, pass_event = self.shortcuts[action]
            if pass_event == False:
                press_func()
            else:
                press_func(event)
            return True
        except KeyError:
            return False

        
# --------------------------------------------------------- effect editors
def connect_filter_widget(widget):
    widget.connect("key-press-event", _filter_widget_keypress_handler)
    children = widget.get_children()
    for child in children:
        child.connect("key-press-event", _filter_widget_keypress_handler)
            
def _filter_widget_keypress_handler(widget, event):
    action = shortcuts.get_shortcut_action(event)
    focus_editor = _get_focus_keyframe_editor(clipeffectseditor.keyframe_editor_widgets)

    if focus_editor != None:
        if focus_editor.get_focus_child() != None:
            if focus_editor.__class__ == keyframeeditor.FilterRectGeometryEditor or \
                focus_editor.__class__ == keyframeeditor.FilterRotatingGeometryEditor or \
                focus_editor.__class__ == keyframeeditor.GeometryNoKeyframes:
                if ((event.keyval == Gdk.KEY_Left) 
                    or (event.keyval == Gdk.KEY_Right)
                    or (event.keyval == Gdk.KEY_Up)
                    or (event.keyval == Gdk.KEY_Down)):
                    focus_editor.arrow_edit(event.keyval, (event.get_state() & Gdk.ModifierType.CONTROL_MASK), (event.get_state() & Gdk.ModifierType.SHIFT_MASK))
                    return True
        if action == 'play_pause':
            _play_pause_action()
            return True
        if action == 'play_pause_loop_marks':
            _play_pause_loop_marks_action()
            return True
        if action == 'prev_frame' or action == 'next_frame':
            if action == 'prev_frame':
                _prev_frame_action(event)
            else:
                _next_frame_action(event)
            return True
        if action == "delete":
            focus_editor.delete_pressed()
            return True
    
    return False


def _get_focus_keyframe_editor(keyframe_editor_widgets):
    if keyframe_editor_widgets == None:
        return None
    for kfeditor in keyframe_editor_widgets:
        if kfeditor.get_focus_child() != None:
           return kfeditor
    return None

# --------------------------------------------------------- compositor editors
def connect_compositor_widget(widget):
    widget.connect("key-press-event", _compositor_widget_keypress_handler)
    children = widget.get_children()
    for child in children:
        child.connect("key-press-event", _compositor_widget_keypress_handler)
        
def _compositor_widget_keypress_handler(widget, event):
    if compositeeditor.keyframe_editor_widgets != None:
        for kfeditor in compositeeditor.keyframe_editor_widgets:
            if kfeditor.get_focus_child() != None:
                if kfeditor.__class__ == keyframeeditor.GeometryEditor or \
                kfeditor.__class__ == keyframeeditor.RotatingGeometryEditor:
                    action = shortcuts.get_shortcut_action(event)
                    if ((event.keyval == Gdk.KEY_Left) 
                        or (event.keyval == Gdk.KEY_Right)
                        or (event.keyval == Gdk.KEY_Up)
                        or (event.keyval == Gdk.KEY_Down)):
                        kfeditor.arrow_edit(event.keyval, (event.get_state() & Gdk.ModifierType.CONTROL_MASK), (event.get_state() & Gdk.ModifierType.SHIFT_MASK))
                        return True
                    if action == 'play_pause':
                        _play_pause_action()
                        return True
                    if action == 'play_pause_loop_marks':
                        _play_pause_loop_marks_action()
                        return True
                    if action == 'prev_frame' or action == 'next_frame':
                        if action == 'prev_frame':
                            _prev_frame_action(event)
                        else:
                            _next_frame_action(event)
                        return True
                    if action == "delete":
                        kfeditor.delete_pressed()
                        return True
    return False

# --------------------------------------------------------- local handler funcs
def _play_pause_action():
    if PLAYER().is_playing():
        monitorevent.stop_pressed()
    else:
        monitorevent.play_pressed()

def _prev_frame_action(event):
    if editorstate.current_is_active_trim_mode() == True:
        trimmodes.left_arrow_pressed((event.get_state() & Gdk.ModifierType.CONTROL_MASK))
    else:    
        _do_arrow_frame_action(-1, event)

def _next_frame_action(event):
    if editorstate.current_is_active_trim_mode() == True:
        trimmodes.right_arrow_pressed((event.get_state() & Gdk.ModifierType.CONTROL_MASK))
    else:    
        _do_arrow_frame_action(1, event)
        
def _do_arrow_frame_action(seek_amount, event):
    prefs = editorpersistance.prefs
    
    if (event.get_state() & Gdk.ModifierType.CONTROL_MASK):
        PLAYER().slowmo_seek_delta(seek_amount)
        return 

    if (event.get_state() & Gdk.ModifierType.SHIFT_MASK):
        seek_amount = seek_amount * prefs.ffwd_rev_shift
    if (event.get_state() & Gdk.ModifierType.LOCK_MASK):
        seek_amount = seek_amount * prefs.ffwd_rev_caps
    PLAYER().seek_delta(seek_amount)

def _next_cut_action():
    if editorstate.current_is_active_trim_mode() == True:
        return

    if editorstate.timeline_visible():
        tline_frame = PLAYER().tracktor_producer.frame()
        cut_frame = current_sequence().find_next_cut_frame(tline_frame) # returns-1 if nor found
        mark_frame = current_sequence().find_next_mark_or_marker(tline_frame) # returns-1 if nor found

        frame = cut_frame
        if mark_frame != -1 and mark_frame < frame:
            frame = mark_frame

        if frame != -1:
            PLAYER().seek_frame(frame)
            if editorpersistance.prefs.center_on_arrow_move == True:
                updater.center_tline_to_current_frame()
    else:
         monitorevent.up_arrow_seek_on_monitor_clip()
     
def _prev_cut_action():
    if editorstate.current_is_active_trim_mode() == True:
        return
     
    if editorstate.timeline_visible():
        tline_frame = PLAYER().tracktor_producer.frame()
        cut_frame = current_sequence().find_prev_cut_frame(tline_frame)
        mark_frame = current_sequence().find_prev_mark_or_marker(tline_frame) # returns-1 if nor found
        
        frame = cut_frame
        if mark_frame != -1 and mark_frame > frame:
            frame = mark_frame

        if frame != -1:
            PLAYER().seek_frame(frame)
            if editorpersistance.prefs.center_on_arrow_move == True:
                updater.center_tline_to_current_frame()  
    else:
         monitorevent.down_arrow_seek_on_monitor_clip()

def _enter_edit_action():
    # We are currently always calling both of these.
    # Check exit of this, e.g. we go from two roll to one roll.
    if editorstate.current_is_active_trim_mode() == True:
        trimmodes.enter_pressed()
            
    if editorstate.EDIT_MODE() == editorstate.MULTI_TRIM:
        multitrimmode.enter_pressed()

def _tline_delete_action():
    if editorstate.EDIT_MODE() == editorstate.KF_TOOL:
        kftoolmode.delete_active_keyframe()
    else:
        # Clip selection and compositor selection are mutually exclusive, 
        # so max one one these will actually delete something.
        tlineaction.splice_out_button_pressed()
        compositormodes.delete_current_selection()

def _play_pause_loop_marks_action():
    if editorstate.current_is_move_mode():
        if PLAYER().is_playing():
            monitorevent.stop_pressed()
        else:
            monitorevent.start_marks_looping()

def _to_start_action():
    if PLAYER().is_playing():
        monitorevent.stop_pressed()
    gui.editor_window.tline_cursor_manager.set_default_edit_tool()
    PLAYER().seek_frame(0)

    if editorstate.timeline_visible():
        tlinewidgets.pos = 0
        updater.repaint_tline()
        updater.update_tline_scrollbar()

def _to_end_action():
    if PLAYER().is_playing():
        monitorevent.stop_pressed()
    gui.editor_window.tline_cursor_manager.set_default_edit_tool()
    PLAYER().seek_end()

    if editorstate.timeline_visible():
        pos = current_sequence().get_length() - 1 - int(float(gui.tline_canvas.widget.get_allocation().width) / float(tlinewidgets.pix_per_frame) * 0.75)
        if pos < 0:
            pos = 0
        tlinewidgets.pos = pos
        updater.repaint_tline()
        updater.update_tline_scrollbar()

def _delete_sequence_action():
    if gui.sequence_list_view.text_rend_1.get_property("editing") == True:
        return
    projectaction.delete_selected_sequence()

def _delete_bin_action():
    if gui.bin_list_view.text_rend_1.get_property("editing") == True:
        return
    projectaction.delete_selected_bin()
    
    
def _delete_log_action():
    # Delete media log event
    if gui.editor_window.media_log_events_list_view.get_focus_child() != None:
        medialog.delete_selected()




"""
    Flowblade Movie Editor is a nonlinear video editor.
    Copyright 2025 Janne Liljeblad.

    This file is part of Flowblade Movie Editor <https://github.com/jliljebl/flowblade/>.

    Flowblade Movie Editor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Flowblade Movie Editor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Flowblade Movie Editor.  If not, see <http://www.gnu.org/licenses/>.
"""
from gi.repository import Gdk, Gtk

import appconsts
import clipeffectseditor
import compositeeditor
import editorstate
import gui
import projectaction
import tlineaction


# ----------------------------------------------------------------------- COPY PASTE ACTION FORWARDING
def cut_action():
    if gui.timeline_has_focus() == False:
        # Try to cut text to clipboard because user pressed CTRL + X.
        if gui.media_list_view.widget.get_focus_child() != None:
            projectaction.cut_media_files()
            return True
                    
        # Try to extract text to clipboard because user pressed CTRL + C.
        copy_source = gui.editor_window.window.get_focus()
        try:
            display = Gdk.Display.get_default()
            cb = Gtk.Clipboard.get_default(display)
            copy_source.get_buffer().cut_clipboard(cb, True)
            return True
        except:# selected widget was not a Gtk.Editable that can provide text to clipboard.
            return False
    else:
        tlineaction.do_timeline_objects_copy(False)
        return True
        
def copy_action():
    if gui.timeline_has_focus() == False:
        filter_kf_editor = _get_focus_keyframe_editor(clipeffectseditor.keyframe_editor_widgets)
        geom_kf_editor = _get_focus_keyframe_editor(compositeeditor.keyframe_editor_widgets)
        if filter_kf_editor != None:
            value = filter_kf_editor.get_copy_kf_value()
            save_data = (appconsts.COPY_PASTE_KEYFRAME_EDITOR_KF_DATA, (value, filter_kf_editor))
            editorstate.set_copy_paste_objects(save_data) 
            return True
        elif geom_kf_editor != None:
            value = geom_kf_editor.get_copy_kf_value() 
            save_data = (appconsts.COPY_PASTE_GEOMETRY_EDITOR_KF_DATA, (value, geom_kf_editor))
            editorstate.set_copy_paste_objects(save_data)
            return True
        else:
            # Try to extract text to clipboard because user pressed CTRL + C.
            copy_source = gui.editor_window.window.get_focus()
            try:
                display = Gdk.Display.get_default()
                cb = Gtk.Clipboard.get_default (display)
                copy_source.get_buffer().copy_clipboard(cb)
                return True
            except:# selected widget was not a Gtk.Editable that can provide text to clipboard.
                return False
    else:
        tlineaction.do_timeline_objects_copy()
        return True

def paste_action():
    if gui.timeline_has_focus() == False:
        copy_paste_object = editorstate.get_copy_paste_objects()
        if copy_paste_object == None:
            _attempt_default_paste()
            return False
        data_type, paste_data = editorstate.get_copy_paste_objects()
        if data_type == appconsts.COPY_PASTE_KEYFRAME_EDITOR_KF_DATA:
            value, kf_editor = paste_data
            kf_editor.paste_kf_value(value)
            return True
        
        elif data_type == appconsts.COPY_PASTE_GEOMETRY_EDITOR_KF_DATA:
            value, geom_editor = paste_data
            geom_editor.paste_kf_value(value)
            return True
        elif data_type == appconsts.CUT_PASTE_MEDIA_ITEMS:
            projectaction.paste_media_files()
            return True
        
        return False
    else:
        tlineaction.do_timeline_objects_paste()
        _attempt_default_paste()
        return True

def _attempt_default_paste():
    # Try to extract text to clipboard because user pressed CTRL + C.
    paste_target = gui.editor_window.window.get_focus()
    try:
        display = Gdk.Display.get_default()
        clipboard = Gtk.Clipboard.get_default(display)
        paste_target.get_buffer().paste_clipboard(clipboard, None, True)
        return True
    except:# selected widget cannot be pasted into
        return False

def _get_focus_keyframe_editor(keyframe_editor_widgets):
    if keyframe_editor_widgets == None:
        return None
    for kfeditor in keyframe_editor_widgets:
        if kfeditor.get_focus_child() != None:
           return kfeditor
    return None
    



"""
    Flowblade Movie Editor is a nonlinear video editor.
    Copyright 2012 Janne Liljeblad.

    This file is part of Flowblade Movie Editor <https://github.com/jliljebl/flowblade/>.

    Flowblade Movie Editor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Flowblade Movie Editor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Flowblade Movie Editor.  If not, see <http://www.gnu.org/licenses/>.
"""

import appconsts
import edit
import resync
import editorpersistance


ONE_ROLL_TRIM_LAST_END = 0
ONE_ROLL_TRIM_START = 1
ONE_ROLL_TRIM_END = 2


def set_child_clip_trim_data(edit_data, edit_tool):
    if edit_data["to_side_being_edited"] == True:
        parent_clip = edit_data["to_clip"]
    else:
        parent_clip = edit_data["from_clip"]
    
    _set_child_clip_data(edit_data, parent_clip)

def set_child_clip_end_drag_data(edit_data, parent_clip):
    _set_child_clip_data(edit_data, parent_clip)
    
def _set_child_clip_data(edit_data, parent_clip):
    child_clip_sync_items = resync.get_child_clips(parent_clip)

    if editorpersistance.prefs.auto_sync_single_childs_on_trim == False:
        edit_data["child_clip_trim_data"] = None
        return

    if child_clip_sync_items == None:
        edit_data["child_clip_trim_data"] = None
        return
    
    if len(child_clip_sync_items) > 1:
        edit_data["child_clip_trim_data"] = None
        # TODO: We dont dual trim when clip has multiple children. Set flag for info window.
        return

    edit_data["child_clip_trim_data"] = child_clip_sync_items[0]

def get_one_roll_sync_edit(edit_data, delta, trim_type):
    if edit_data["child_clip_trim_data"] == None:
        return None

    if editorpersistance.prefs.auto_sync_single_childs_on_trim == False:
        return None
        
    child_clip, child_track = edit_data["child_clip_trim_data"]

    if trim_type == ONE_ROLL_TRIM_LAST_END:
        if delta > 0 and child_clip.get_length() - child_clip.clip_out < delta:
            return None
        if delta < 0 and child_clip.clip_out - child_clip.clip_in < abs(delta):
            return None
        if delta == 0:
            return None

        data = {"track":child_track,
                "index":child_track.clips.index(child_clip),
                "clip":child_clip,
                "delta":delta,
                "undo_done_callback":None,
                "first_do":False}
        action = edit.trim_last_clip_end_action(data)
        return action

    elif trim_type == ONE_ROLL_TRIM_START:
        if delta > 0 and child_clip.clip_out - child_clip.clip_in < delta:
            return None
        if delta < 0 and child_clip.clip_in < abs(delta):
            return None
        if delta == 0:
            return None
        data = {"track":child_track,
                "index":child_track.clips.index(child_clip),
                "clip":child_clip,
                "delta":delta,
                "undo_done_callback":None,
                "first_do":False}
        action = edit.trim_start_action(data)
        return action

    elif trim_type == ONE_ROLL_TRIM_END:
        if delta > 0 and child_clip.get_length() - child_clip.clip_out < delta:
            return None
        if delta < 0 and child_clip.clip_out - child_clip.clip_in < abs(delta):
            return None
        if delta == 0:
            return None
        data = {"track":child_track,
                "index":child_track.clips.index(child_clip),
                "clip":child_clip,
                "delta":delta,
                "undo_done_callback":None,
                "first_do":False}
        action = edit.trim_end_action(data)
        return action
    return None

def get_two_roll_sync_edits(edit_data):
    if editorpersistance.prefs.auto_sync_single_childs_on_trim == False:
        return None
        
    delta = edit_data["delta"]
    from_clip_sync_items = resync.get_child_clips(edit_data["from_clip"])

    if from_clip_sync_items == None or len(from_clip_sync_items) > 1:
        from_clip = from_track = None
    else:
        from_clip, from_track = from_clip_sync_items[0]
        from_index = from_track.clips.index(from_clip) 

    to_clip_sync_items = resync.get_child_clips(edit_data["to_clip"])

    if to_clip_sync_items == None or len(to_clip_sync_items) > 1:
        to_clip = to_track = None
    else:
        to_clip, to_track = to_clip_sync_items[0]
        to_index = to_track.clips.index(to_clip)
        
    
    # CASE: to and from clip on same track immediately one after another
    if to_track != None and from_track != None and to_track == from_track and to_index == from_index + 1:
        data = {"track":to_track,
                "index":to_index,
                "from_clip":from_clip,
                "to_clip":to_clip,
                "delta":edit_data["delta"],
                "edit_done_callback": None, # we don't do callback needing this
                "cut_frame": edit_data["cut_frame"],
                "to_side_being_edited":None, # we don't do callback needing this
                "non_edit_side_blank":False,
                "first_do":False}  # no callback

        sync_trim_legal = check_two_roll_trim_lagality(data)
        if sync_trim_legal == False:
            return None

        action = edit.tworoll_trim_action(data)
        return [action]
    
    actions = []
    
    if from_clip != None:
        if from_index + 1 < len(from_track.clips):
            to_clip_one = from_track.clips[from_index + 1]
            non_edit_side_blank = (to_clip_one.is_blank == True)
            data = {"track":from_track,
                    "index":from_index + 1,
                    "from_clip":from_clip,
                    "to_clip":to_clip_one,
                    "delta":edit_data["delta"],
                    "edit_done_callback": None, # we don't do the callback needing this
                    "cut_frame": edit_data["cut_frame"],
                    "to_side_being_edited":None, # we don't do the callback needing this
                    "non_edit_side_blank":non_edit_side_blank,
                    "first_do":False}  # no callback
            
            if check_two_roll_trim_lagality(data) == True:
                action = edit.tworoll_trim_action(data)
                actions.append(action)
        else:
            do_edit = True
            if delta > 0 and from_clip.get_length() - from_clip.clip_out < delta:
                do_edit = False
            if delta < 0 and from_clip.clip_out - from_clip.clip_in < abs(delta):
                do_edit = False
            if delta == 0:
                do_edit = False
            
            if do_edit == True:
                data = {"track":from_track,
                        "index":from_index,
                        "clip":from_clip,
                        "delta":delta,
                        "undo_done_callback":None,
                        "first_do":False}
                action = edit.trim_last_clip_end_action(data)
                actions.append(action)
                
    if to_clip != None and to_index > 0:
        from_clip_here = to_track.clips[to_index - 1]
        non_edit_side_blank = (from_clip_here.is_blank == True)
        data = {"track":to_track,
                "index":to_index,
                "from_clip":from_clip_here,
                "to_clip":to_clip,
                "delta":edit_data["delta"],
                "edit_done_callback": None, # we don't do the callback needing this
                "cut_frame": edit_data["cut_frame"],
                "to_side_being_edited":None, # we don't do the callback needing this
                "non_edit_side_blank":non_edit_side_blank,
                "first_do":False}  # no callback
        
        if check_two_roll_trim_lagality(data) == True:
            action = edit.tworoll_trim_action(data)
            actions.append(action)
            
    if len(actions) == 0:
        return None
    else:
        return actions

def check_two_roll_trim_lagality(edit_data):
    cut_frame = edit_data["cut_frame"]
    from_clip = edit_data["from_clip"]
    to_clip = edit_data["to_clip"]
    delta = edit_data["delta"]
    
    # Trim_limits frames here are TIMELINE frames, not CLIP frames
    trim_limits = {}

    trim_limits["from_start"] = cut_frame - (from_clip.clip_out - from_clip.clip_in)
    from_length = from_clip.get_length()
    trim_limits["from_end"] = cut_frame - from_clip.clip_out + from_length - 2 # -1 incl, -1 leave one frame, == -2

    if from_clip.is_blanck_clip:
        trim_limits["from_end"]  = 10000000

    trim_limits["to_start"] = cut_frame - to_clip.clip_in
    to_length = to_clip.get_length()
    trim_limits["to_end"] = cut_frame + (to_clip.clip_out - to_clip.clip_in) #- to_clip.clip_in + to_length - 1 # - 1, leave one frame
    if to_clip.is_blanck_clip:
        trim_limits["to_start"] = 0

    if trim_limits["from_start"] > trim_limits["to_start"]:
        trim_limits["both_start"] = trim_limits["from_start"]
    else:
        trim_limits["both_start"] = trim_limits["to_start"]
        
    if trim_limits["to_end"] < trim_limits["from_end"]:
        trim_limits["both_end"] = trim_limits["to_end"]
    else:
        trim_limits["both_end"] = trim_limits["from_end"]

    start_available_range_length = abs(cut_frame - trim_limits["both_start"])
    end_available_range_length = abs(cut_frame - trim_limits["both_end"])

    if start_available_range_length < abs(delta):
        return False
    if end_available_range_length < abs(delta):
        return False

    return True

def get_slide_trim_sync_edit(edit_data):
    if editorpersistance.prefs.auto_sync_single_childs_on_trim == False:
        return None

    clip_sync_items = resync.get_child_clips(edit_data["clip"])

    if clip_sync_items == None or len(clip_sync_items) > 1:
        return None
    else:
        sync_clip, sync_track = clip_sync_items[0]
        sync_index = sync_track.clips.index(sync_clip) 

    delta = edit_data["delta"]

    if sync_clip.clip_in + delta < 0:
        return None
    if sync_clip.clip_out + delta >= edit_data["clip"].get_length():
        return None

    data = {"track":sync_track,
            "index":sync_index,
            "clip":sync_clip,
            "delta":delta,
            "first_do_callback":None, # not used because first_do == False
            "start_frame_being_viewed":None, # not used because first_do == False
            "first_do":False} 
    action = edit.slide_trim_action(data)

    return action



"""
    Flowblade Movie Editor is a nonlinear video editor.
    Copyright 2025 Janne Liljeblad.

    This file is part of Flowblade Movie Editor <https://github.com/jliljebl/flowblade/>.

    Flowblade Movie Editor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Flowblade Movie Editor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Flowblade Movie Editor.  If not, see <http://www.gnu.org/licenses/>.
"""

from gi.repository import Gtk, Gdk
                        

class SimpleStateEvent:
    
    def __init__(self):
        self.event = Gtk.get_current_event()
        unknow_val, self.state = self.event.get_state()
        
    def get_state(self):
        return self.state

    def get_gdk_event(self):
        return self.event

class ScrollEvent:
    
    def __init__(self, dx, dy):
        if dy < 0.0:
            self.direction = Gdk.ScrollDirection.UP
        elif dy > 0.0:
            self.direction = Gdk.ScrollDirection.DOWN
        elif dx > 0.0:
            self.direction = Gdk.ScrollDirection.RIGHT
        else:
            self.direction = Gdk.ScrollDirection.LEFT

        event = Gtk.get_current_event()
        unknow_val, self.state = event.get_state()
        
    def get_state(self):
        return self.state


class ButtonEvent:
    
    def __init__(self, n_press, x, y):
    
        self.x = x
        self.y = y 
        self.n_press = n_press

        event = Gtk.get_current_event()
        unknown_val, self.state = event.get_state()
    
        self.type = event.type
        self.button = event.button.button

    def get_state(self):
        return self.state

class KeyPressEvent:

    def __init__(self, event, keyval, keycode, state):
        #event
        self.keyval = keyval
        self.keycode = keycode
        self.state = state
 
    def get_state(self):
        return self.state

class KeyPressEventAdapter:
    
    def __init__(self, widget, callback, user_data=None, capture=False):
        self.controller = Gtk.EventControllerKey.new(widget)
        if capture != False:
            self.controller.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        self.controller.connect("key-pressed", self.pressed_event)
        self.widget = widget
        self.callback = callback
        self.user_data = user_data 
    
    def pressed_event(self, event, keyval, keycode, state):
        gdk_event = KeyPressEvent(event, keyval, keycode, state)
        if self.user_data == None:
            self.callback(self.widget, gdk_event)
        else:
            self.callback(self.widget, gdk_event, *user_data)

